from pycparser import c_ast as ast

from .Node import Node
from .VariableNode import VariableNode
from .Collection import Collection


class VariableNodeCollection(Node, Collection):
    def __init__(self, top_node):
        self.variables: list[VariableNode] = []
        for nested in top_node:
            if VariableNode.is_type(nested):
                self.variables.append(VariableNode(nested))
            elif VariableNode.is_type(nested.type):
                self.variables.append(VariableNode(nested.type))

    @staticmethod
    def is_type(node):
        return isinstance(node, ast.FileAST)



import sys

from pycparser import c_ast as ast

from .Node import Node


class TypeNodeMap(Node):
    _c2python_types_map = {
        'void': 'None',
        'char': 'int',
        'signed char': 'int',
        'unsigned char': 'int',
        'short': 'int',
        'short int': 'int',
        'signed short': 'int',
        'signed short int': 'int',
        'unsigned short': 'int',
        'unsigned short int': 'int',
        'int': 'int',
        'signed': 'int',
        'signed int': 'int',
        'unsigned': 'int',
        'unsigned int': 'int',
        'long': 'int',
        'long int': 'int',
        'signed long': 'int',
        'signed long int': 'int',
        'unsigned long': 'int',
        'unsigned long int': 'int',
        'long long': 'int',
        'long long int': 'int',
        'signed long long': 'int',
        'signed long long int': 'int',
        'unsigned long long': 'int',
        'unsigned long long int': 'int',
        'float': 'float',
        'double': 'float',
        'long double': 'float'
    }

    def __init__(self, top_node):
        self.map = {}
        for nested in top_node:
            if not isinstance(nested, ast.Typedef):
                continue
            nested.type, _ = Node._skip_ptr_decls(nested.type)
            if not isinstance(nested.type, ast.TypeDecl):
                continue
            if isinstance(nested.type.type, ast.IdentifierType):
                self._update_types(nested.type.declname, " ".join(nested.type.type.names))
            elif isinstance(nested.type.type, ast.Struct):
                self._update_types(nested.type.declname, nested.type.type.name)

    @staticmethod
    def is_type(node):
        return isinstance(node, ast.FileAST)

    def __getitem__(self, item):
        if item in self.map:
            return self.__getitem__(self.map[item])
        if item in self._c2python_types_map:
            return self._c2python_types_map[item]
        return item

    def _update_types(self, left, right):
        if left in self.map.keys():
            sys.exit("Found duplicated(second definition?) type '" + left + "'")
        self.map[left] = right



from pycparser import c_ast as ast

from .Node import Node


class VariableNode(Node):
    def __init__(self, node):
        node, _ = self._skip_ptr_decls(node.type)
        self.name = node.declname
        self.type = " ".join(node.type.names)

    @staticmethod
    def is_type(node):
        return isinstance(node, ast.Decl) and (isinstance(node.type, ast.TypeDecl) or isinstance(node.type, ast.PtrDecl))


from pycparser import c_ast as ast

from .Node import Node


class StructNodeMember(Node):
    def __init__(self, type_decl_node):
        self.name = type_decl_node.declname
        self.type = " ".join(type_decl_node.type.names)

    @staticmethod
    def is_type(decl_node):
        return isinstance(decl_node.type, ast.IdentifierType)



from pycparser import c_ast as ast

from .Node import Node
from .FunctionNodeParameter import FunctionNodeParameter


class FunctionNode(Node):
    def __init__(self, node):
        node = node.type
        self.name = node.type.declname
        self.return_type = " ".join(node.type.type.names)
        self.params = []
        for decl in node.args.params:
            if FunctionNodeParameter.is_type(decl):
                param = FunctionNodeParameter(decl)
                self.params.append(param)

    @staticmethod
    def is_type(decl_node):
        return (isinstance(decl_node, ast.Decl) or isinstance(decl_node, ast.PtrDecl)) and \
            isinstance(decl_node.type, ast.FuncDecl)

    def param_names(self):
        return [param.name for param in self.params]



import os

from jinja2 import Environment, FileSystemLoader

from ..Collector import Collector
from ..Driver import Driver


class MockDriver(Driver):
    _templ_dir = os.path.abspath(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates')
    )
    _env = Environment(loader=FileSystemLoader(_templ_dir), autoescape=True, keep_trailing_newline=True)

    @classmethod
    async def compose(cls, collector: Collector, output_dir: str, license_text: str) -> tuple[str, str]:
        if not os.path.isdir(output_dir):
            raise ValueError(f"Directory '{output_dir}' not found.")

        pkg_dir = os.path.join(output_dir, collector.python_package_name)
        os.mkdir(pkg_dir)

        init_str = cls._env.get_template("__init__.py.j2").render()
        init_path = os.path.join(pkg_dir, "__init__.py")
        with open(init_path, 'w') as file_obj:
            file_obj.write(init_str)

        base_model_str = cls._env.get_template("BaseModel.py.j2").render()
        base_model_path = os.path.join(pkg_dir, "BaseModel.py")
        with open(base_model_path, 'w') as file_obj:
            file_obj.write(base_model_str)

        model_str = cls._env.get_template("Model.py.j2").render(
            model_name=collector.model_name,
            input_map=collector.input_members,
            output_map=collector.output_members,
            time_step=collector.time_step
        )
        model_path = os.path.join(pkg_dir, "Model.py")
        with open(model_path, 'w') as file_obj:
            file_obj.write(model_str)

        pyproject_str = cls._env.get_template("pyproject.toml.j2").render(
            model_version=collector.model_version,
            package_name=collector.python_package_name,
            model_name=collector.model_name
        )
        pyproject_path = os.path.join(output_dir, "pyproject.toml")
        with open(pyproject_path, 'w') as file_obj:
            file_obj.write(pyproject_str)

        license_path = os.path.join(output_dir, "LICENSE")
        with open(license_path, 'w') as file_obj:
            file_obj.write(license_text)

        readme_str = cls._env.get_template("README.md.j2").render()
        readme_path = os.path.join(output_dir, "README.md")
        with open(readme_path, 'w') as file_obj:
            file_obj.write(readme_str)

        return collector.python_package_name, cls.wrapper_class_name



from pycparser import c_ast as ast

from .Node import Node
from .StructNode import StructNode
from .Collection import Collection


class StructNodeCollection(Node, Collection):
    def __init__(self, top_node):
        self.structs: dict[str, StructNode] = {}
        for nested in top_node:
            if StructNode.is_type(nested):
                self.add(StructNode(nested))
            elif StructNode.is_type(nested.type):
                self.add(StructNode(nested.type))

    @staticmethod
    def is_type(decl_node):
        return isinstance(decl_node, ast.FileAST)

    def add(self, struct: StructNode):
        if struct.name in self.structs.keys():
            raise ValueError(f"Found duplicate struct name: '{struct.name}'")
        self.structs[struct.name] = struct

    def __getitem__(self, item):
        return self.structs[item]

    def __contains__(self, item):
        return item in self.structs.keys()



import asyncio
import multiprocessing
import os
import random
import shutil
import sys
from concurrent.futures import ProcessPoolExecutor
from importlib import import_module

from .Collector import Collector


async def test_model_integrity(
        search_path: str,
        pkg_name: str,
        class_name: str,
        collector: Collector,
        test_callback: bool = False,
):
    loop = asyncio.get_event_loop()
    pool = ProcessPoolExecutor(max_workers=1)
    manager = multiprocessing.Manager()
    queue = manager.Queue(-1)

    await loop.run_in_executor(
        pool,
        execute,
        queue,
        search_path,
        pkg_name,
        class_name,
        collector.input_members.keys(),
        collector.output_members.keys(),
        test_callback
    )
    if not queue.empty():
        raise RuntimeError(queue.get())
    delete_cache(search_path)


def delete_cache(root_dir):
    for root, dirs, files in os.walk(root_dir):
        if "__pycache__" in dirs:
            shutil.rmtree(os.path.join(root, "__pycache__"))


def execute(
        queue: multiprocessing.Queue,
        search_path: str,
        pkg_name: str,
        class_name: str,
        input_vars: list[str],
        output_vars: list[str],
        test_callback: bool
):
    current_time = 0
    prev_time = 0
    callback_called = False
    current_input = {}
    current_output = {}

    def callback(time: float, in_data: dict, out_data: dict):
        nonlocal callback_called
        callback_called = True
        if set(in_data.keys()) != set(input_vars):
            raise RuntimeError("Input vars do not match!")
        if set(out_data.keys()) != set(output_vars):
            raise RuntimeError("Output vars do not match!")
        if time != current_time:
            raise RuntimeError("Time does not match")
        if in_data != current_input:
            raise RuntimeError("Current input does not match")
        if out_data != current_output:
            raise RuntimeError("Current output does not match")

    def collect_data():
        nonlocal current_time
        current_time = getattr(instance, "time")
        input_data = getattr(instance, "input")
        output_data = getattr(instance, "output")
        for val in input_vars:
            current_input[val] = getattr(input_data, val)
        for val in output_vars:
            current_output[val] = getattr(output_data, val)

    def set_input():
        input_data = getattr(instance, "input")
        for val in input_vars:
            setattr(input_data, val, random.uniform(0.0, 1000.0))

    try:
        sys.path.append(search_path)

        model_module = import_module(pkg_name)
        model_class = getattr(model_module, class_name)

        instance = model_class(callback) if test_callback else model_class()
        for _ in range(100):
            set_input()
            collect_data()
            instance.step()
            if test_callback:
                if not callback_called:
                    raise RuntimeError("Callback is not called")
                callback_called = False
            if prev_time == getattr(instance, "time"):
                raise RuntimeError("Time does not increment")
            prev_time = getattr(instance, "time")
        collect_data()
        instance.terminate()
    except Exception as ex:
        queue.put(str(ex))



import abc

from .Collector import Collector


class Driver(abc.ABC):
    wrapper_class_name = "Model"

    @classmethod
    @abc.abstractmethod
    async def compose(cls, architect: Collector, output_dir: str, license_text: str) -> tuple[str, str]:
        pass



from .Node import Node
from .StructNodeCollection import StructNodeCollection
from .FunctionNodeCollection import FunctionNodeCollection
from .VariableNodeCollection import VariableNodeCollection
from .TypeNodeMap import TypeNodeMap
